@@ -4,6 +4,7 @@
  * Copyright (c) 2000-2006 Anton Altaparmakov
  * Copyright (c) 2002-2009 Szabolcs Szakacsits
  * Copyright (c) 2004-2005 Richard Russon
+ * Copyright (c) 2010      Jean-Pierre Andre
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -65,6 +66,7 @@
 #include "logfile.h"
 #include "dir.h"
 #include "logging.h"
+#include "cache.h"
 #include "misc.h"
 
 const char *ntfs_home = 
@@ -197,8 +199,10 @@ static int __ntfs_volume_release(ntfs_volume *v)
 			ntfs_error_set(&err);
 	}
 
+	ntfs_free_lru_caches(v);
 	free(v->vol_name);
 	free(v->upcase);
+	if (v->locase) free(v->locase);
 	free(v->attrdef);
 	free(v);
 
@@ -485,7 +489,14 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 	
 	ntfs_upcase_table_build(vol->upcase,
 			vol->upcase_len * sizeof(ntfschar));
+	/* Default with no locase table and case sensitive file names */
+	vol->locase = (ntfschar*)NULL;
+	NVolSetCaseSensitive(vol);
 	
+		/* by default, all files are shown and not marked hidden */
+	NVolSetShowSysFiles(vol);
+	NVolSetShowHidFiles(vol);
+	NVolClearHideDotFiles(vol);
 	if (flags & MS_RDONLY)
 		NVolSetReadOnly(vol);
 	
@@ -525,6 +536,7 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 				"%s\n", strerror(errno));
 	
 	/* We now initialize the cluster allocator. */
+	vol->full_zones = 0;
 	mft_zone_size = vol->nr_clusters >> 3;      /* 12.5% */
 
 	/* Setup the mft zone. */
@@ -565,9 +577,9 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 	 * respective zone.
 	 */
 	vol->data1_zone_pos = vol->mft_zone_end;
-	ntfs_log_debug("data1_zone_pos = 0x%llx\n", vol->data1_zone_pos);
+	ntfs_log_debug("data1_zone_pos = %lld\n", (long long)vol->data1_zone_pos);
 	vol->data2_zone_pos = 0;
-	ntfs_log_debug("data2_zone_pos = 0x%llx\n", vol->data2_zone_pos);
+	ntfs_log_debug("data2_zone_pos = %lld\n", (long long)vol->data2_zone_pos);
 
 	/* Set the mft data allocation position to mft record 24. */
 	vol->mft_data_pos = 24;
@@ -758,6 +770,70 @@ out:
 	return errno ? -1 : 0;
 }
 
+/*
+ *		Make sure a LOGGED_UTILITY_STREAM attribute named "$TXF_DATA"
+ *	on the root directory is resident.
+ *	When it is non-resident, the partition cannot be mounted on Vista
+ *	(see http://support.microsoft.com/kb/974729)
+ *
+ *	We take care to avoid this situation, however this can be a
+ *	consequence of having used an older version (including older
+ *	Windows version), so we had better fix it.
+ *
+ *	Returns 0 if unneeded or successful
+ *		-1 if there was an error, explained by errno
+ */
+
+static int fix_txf_data(ntfs_volume *vol)
+{
+	void *txf_data;
+	s64 txf_data_size;
+	ntfs_inode *ni;
+	ntfs_attr *na;
+	int res;
+
+	res = 0;
+	ntfs_log_debug("Loading root directory\n");
+	ni = ntfs_inode_open(vol, FILE_root);
+	if (!ni) {
+		ntfs_log_perror("Failed to open root directory");
+		res = -1;
+	} else {
+		/* Get the $TXF_DATA attribute */
+		na = ntfs_attr_open(ni, AT_LOGGED_UTILITY_STREAM, TXF_DATA, 9);
+		if (na) {
+			if (NAttrNonResident(na)) {
+				/*
+				 * Fix the attribute by truncating, then
+				 * rewriting it.
+				 */
+				ntfs_log_debug("Making $TXF_DATA resident\n");
+				txf_data = ntfs_attr_readall(ni,
+						AT_LOGGED_UTILITY_STREAM,
+						TXF_DATA, 9, &txf_data_size);
+				if (txf_data) {
+					if (ntfs_attr_truncate(na, 0)
+					    || (ntfs_attr_pwrite(na, 0,
+						 txf_data_size, txf_data)
+							!= txf_data_size))
+						res = -1;
+					free(txf_data);
+				}
+			if (res)
+				ntfs_log_error("Failed to make $TXF_DATA resident\n");
+			else
+				ntfs_log_error("$TXF_DATA made resident\n");
+			}
+			ntfs_attr_close(na);
+		}
+		if (ntfs_inode_close(ni)) {
+			ntfs_log_perror("Failed to close root");
+			res = -1;
+		}
+	}
+	return (res);
+}
+
 /**
  * ntfs_device_mount - open ntfs volume
  * @dev:	device to open
@@ -1043,9 +1119,9 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 				goto error_exit;
 			
 			for (j = 0; j < (s32)u; j++) {
-				ntfschar uc = le16_to_cpu(vname[j]);
+				u16 uc = le16_to_cpu(vname[j]);
 				if (uc > 0xff)
-					uc = (ntfschar)'_';
+					uc = (u16)'_';
 				vol->vol_name[j] = (char)uc;
 			}
 			vol->vol_name[u] = '\0';
@@ -1109,6 +1185,9 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 				goto error_exit;
 		}
 	}
+		/* make $TXF_DATA resident if present on the root directory */
+	if (!NVolReadOnly(vol) && fix_txf_data(vol))
+		goto error_exit;
 
 	return vol;
 io_error_exit:
@@ -1124,6 +1203,58 @@ error_exit:
 	return NULL;
 }
 
+/*
+ *		Set appropriate flags for showing NTFS metafiles
+ *	or files marked as hidden.
+ *	Not set in ntfs_mount() to avoid breaking existing tools.
+ */
+
+int ntfs_set_shown_files(ntfs_volume *vol,
+			BOOL show_sys_files, BOOL show_hid_files,
+			BOOL hide_dot_files)
+{
+	int res;
+
+	res = -1;
+	if (vol) {
+		NVolClearShowSysFiles(vol);
+		NVolClearShowHidFiles(vol);
+		NVolClearHideDotFiles(vol);
+		if (show_sys_files)
+			NVolSetShowSysFiles(vol);
+		if (show_hid_files)
+			NVolSetShowHidFiles(vol);
+		if (hide_dot_files)
+			NVolSetHideDotFiles(vol);
+		res = 0;
+	}
+	if (res)
+		ntfs_log_error("Failed to set file visibility\n");
+	return (res);
+}
+
+/*
+ *		Set ignore case mode
+ */
+
+int ntfs_set_ignore_case(ntfs_volume *vol)
+{
+	int res;
+
+	res = -1;
+	if (vol && vol->upcase) {
+		vol->locase = ntfs_locase_table_build(vol->upcase,
+					vol->upcase_len);
+		if (vol->locase) {
+			NVolClearCaseSensitive(vol);
+			res = 0;
+		}
+	}
+	if (res)
+		ntfs_log_error("Failed to set ignore_case mode\n");
+	return (res);
+}
+
 /**
  * ntfs_mount - open ntfs volume
  * @name:	name of device/file to open
@@ -1166,7 +1297,8 @@ ntfs_volume *ntfs_mount(const char *name __attribute__((unused)),
 		int eo = errno;
 		ntfs_device_free(dev);
 		errno = eo;
-	}
+	} else
+		ntfs_create_lru_caches(vol);
 	return vol;
 #else
 	/*
@@ -1429,7 +1561,7 @@ error_exit:
  *
  * Return 0 if successful and -1 if not with errno set to the error code.
  */
-int ntfs_volume_write_flags(ntfs_volume *vol, const u16 flags)
+int ntfs_volume_write_flags(ntfs_volume *vol, const le16 flags)
 {
 	ATTR_RECORD *a;
 	VOLUME_INFORMATION *c;
@@ -1562,3 +1694,30 @@ int ntfs_set_locale(void)
 	return 0;
 }
 
+/*
+ *		Feed the counts of free clusters and free mft records
+ */
+
+int ntfs_volume_get_free_space(ntfs_volume *vol)
+{
+	ntfs_attr *na;
+	int ret;
+
+	ret = -1; /* default return */
+	vol->free_clusters = ntfs_attr_get_free_bits(vol->lcnbmp_na);
+	if (vol->free_clusters < 0) {
+		ntfs_log_perror("Failed to read NTFS $Bitmap");
+	} else {
+		na = vol->mftbmp_na;
+		vol->free_mft_records = ntfs_attr_get_free_bits(na);
+
+		if (vol->free_mft_records >= 0)
+			vol->free_mft_records += (na->allocated_size - na->data_size) << 3;
+
+		if (vol->free_mft_records < 0)
+			ntfs_log_perror("Failed to calculate free MFT records");
+		else
+			ret = 0;
+	}
+	return (ret);
+}
